%
% WebLS -- A Web-Based Inference Engine and CGI Interface
% Version 1.0
% Copyright (c) 1996 by Amzi! inc. All Rights Reserved
%
%
% See license.htm for Amzi! License Agreement
% 
% Requirements:
%   Amzi! Prolog + Logic Server 3.3 Mar96
%   Win-CGI Compliant Web Server for Windows NT or 95
%
% Architecture:
%
% This allows the user of Amzi! Prolog + Logic Server as a Windows
% CGI script.  It can be used to answer tech support questions,
% diagnose problems, give configuration advice, act as intelligent
% site guide and give sales advice.
%
% This is a work in progress.  Check our web site for the latest and
% greatest version, http://www.amzi.com
%
% There are 3 parts:
%   The Prolog CGI Interface
%   The Prolog Inference Engine (in amzieng.pro)
%   The Logic-Base (in logicbse.pro)
%
% The Prolog CGI Interface is a collection of routines that depend upon C
% functions in amzicgi.exe to process CGI input and output.  There are routines
% to read the INI file, write the output file, keep a log, etc.  Generally
% you will not change these.  All the cgi routines have names beginning with 
% 'cgi'.
%
% This is a self-contained module except that it expects you to provide the
% definition for these 2 predicates:
%    cgiStart
%    cgiMain
%
% This interface is used by the Prolog Inference Engine.  Because CGI is a 
% transaction based interface, the Prolog Engine needs to maintain state
% information across transactions.  This is done by hiding the name of
% the state file in the submit buttons.  
%
% The Engine also wants to minimize the number of transactions needed to
% reach an answer.  To do this, first it sees if any rule can be
% executed with the information already gathered.  And if that fails,
% it gathers as much information as possible for the current hypothesis.
%
% The logic-base is a simple set of Prolog rules that check various facts,
% and output solutions.  Although in this example the answers are hard-
% coded, support is provided in the Prolog CGI Interface to redirect 
% the user to an existing HTML document, or to output an existing
% document.  The latter strategy is used to collect the initial problem
% report from the user.
%

:- op(900, fy, [?, bug]).

% ------------------------------------------------------------------- %
%                        Prolog CGI Interface                         %
% ------------------------------------------------------------------- %

%
% main
%
% The main program.  cgiStart is used to setup the log file and any items 
% needed for debugging your script (usually a dummy INI file).  cgiMain
% is the script itself, and is provided by you.  cgiInit and cgiFinish
% setup the environment for script writing as follows:
%   They open and close the output file
%   They read CGI parameters from the INI input file and assert them
%
main :-
   cgiStart,
   cgiInit,
   catch((cgiMain, cgiErrorCheck), X, cgiError(X)),
   cgiFinish.
main :-
   cgiError($cgiMain/0 failed$),
   cgiFinish.

%
% cgiInit
%
% Opens the output file and log file and gets the values of various CGI 
% variables
%
cgiInit :-
   % Load the environment variables from the INI file (if present)
   % For normal CGI the values have already been asserted
   cgiInitINI,

   % Open the output file
   cgiInitOutput.

cgiInitINI :-
   % Get the input file.  This was asserted by amzicgi.c/exe for Win-CGI
   % For standard CGI this will fail, and we don't read any INI values
   system('Input File', INIFile),
   !,

   % Assert all the system and CGI facts
   cgiGetInputValues($CGI$, cgi, INIFile),
   cgiGetInputValues($System$, system, INIFile),
   cgiGetInputValues($Form Literal$, fact, INIFile),
   cgiGetInputValues($Form External$, fact, INIFile),
   cgiGetInputValues($Form File$, fact, INIFile),
   cgiGetInputValues($Form Huge$, fact, INIFile),
   cgiGetInputValues($Accept$, accept, INIFile),
   cgiGetInputValues($Extra Headers$, extraHeaders, INIFile).
cgiInitINI.

cgiInitOutput :-
   % Get the output file.  This is in the INI file for Win-CGI and 
   % fails for normal CGI.
   system('Output File', OF),
   !,
   cgiOpen(OF).
cgiInitOutput :-
   cgiOpen($$).   

%
% cgiFinish
%
% Closes the output file and log file so we can exit
%
cgiFinish :-
   cgiClose,
   cgiCloseLog.
cgiFinish.

%
% err_handle/3
%
% These are standard Prolog error handlers.  All errors are caught and saved
% in the dynamic database.  For read errors (300's) the offending text is
% saved with a marker <- NEAR HERE -> where the problem is.
%
err_handle(300, Term, 0) :-
   err_msg(300, MsgTextL),
   string_list(MsgText, MsgTextL),
   err_read(ReadTextL, Index),
   string_list(ReadText, ReadTextL),
   sub_string(ReadText, 1, Index, FirstPart),
   I2 is Index + 1,
   sub_string(ReadText, I2, _, LastPart),
   strcat(FirstPart, $ <-NEAR HERE-> $, R1),
   strcat(R1, LastPart, R2),
   string_list(R2, ListIn),
   cgiExpandChars(ListIn, ListOut),
   string_list(ReadTextX, ListOut),
   assertz(cgiErrorData(300, MsgText, ReadTextX)).
err_handle(Num, Term, 0) :-
   err_msg(Num, MsgTextL),
   string_list(MsgText, MsgTextL),
   string_term(TermS, Term),
   assertz(cgiErrorData(Num, MsgText, TermS)).

%
% cgiErrorCheck
%
% Checks for asserted errors, which are caught by err_handle (above).  If any
% are found, throw an error to abort out of wherever we were.
%
cgiErrorCheck :-
   cgiErrorData(_, _, _),
   throw($Prolog Engine Execution Errors$).
cgiErrorCheck.

%
% cgiError(Message)
%
% Handles thrown errors--rewinds the output file and sends out the error 
% message in an HTML page.
%
cgiError(Message) :-
   cgiLog(Message),
   cgiErrorHeader,
   cgiSend(Message),
   cgiSend($<P>$), 
   cgiErrorReport,
   cgiSend($<P>$), 
   outputNote(systemError),
   cgiSend($<P><HR><CENTER>$), 
   (system('Log File URL', LFU) -> 
      cgiSend([$<BR>Press <A HREF="$, LFU, $">here</A> (then your browser's reload button) for Logic-Base Debugging Trace$]) 
      ; 
      true
   ),
   cgiSend($</CENTER></BODY></HTML>$),
   cgiSend($$).
% Never fail
cgiError(_).

cgiErrorHeader :-
   cgiRewind,
   % The following lines cause unwanted results with some servers (notably Apache)
%   cgiSend($HTTP/1.0 500 Internal Error$),
%   cgi('Server Software', Server),
%   cgiSend([$Server: $, Server]),
%   gmtime_str(Time),
%   cgiSend([$Date: $, Time]),
   cgiSend($Content-type: text/html$),
   cgiSend($$),
   cgiSend([$<HTML><HEAD><TITLE>WebLS Error</TITLE></HEAD>$]),
   cgiSend([$<BODY bgcolor=#FFB5DA text=#000000><H1>WebLS Error</H1>$]).

cgiErrorReport :-
   cgiErrorData(Num, MsgText, MsgInfo),
   cgiErrorOut(Num, MsgText, MsgInfo),
   fail.
cgiErrorReport :-
   retractall(cgiErrorData(_, _, _)).

cgiErrorOut(300, MsgText, ReadText) :-
   !, 
   cgiSend([$Prolog Error #300: $, MsgText, $<BR>In logic-base:<BR>$, ReadText, $<P>$]).
cgiErrorOut(Num, MsgText, MsgInfo) :-
   cgiSend([$Prolog Error #$, Num, $: $, MsgText, $<BR>$, MsgInfo, $<P>$]).

%
% cgiExpandChars(InputList, OutputList)
%
% Given a character list, these clauses expand special characters into their
% proper HTML representations, e.g. > becomes &gt;  This is used by the
% error handler when reporting syntax (read) errors as the buffer contains
% parts of the logic-base which include HTML tags which are not to be
% interpreted, simply displayed.
%
cgiExpandChars([], []).
cgiExpandChars([`> | T1], [`&,`g,`t,`; | T2]) :- !, cgiExpandChars(T1, T2).
cgiExpandChars([`< | T1], [`&,`l,`t,`; | T2]) :- !, cgiExpandChars(T1, T2).
cgiExpandChars([`& | T1], [`&,`a,`m,`p,`; | T2]) :- !, cgiExpandChars(T1, T2).
cgiExpandChars([`" | T1], [`&,`q,`u,`o,`t,`; | T2]) :- !, cgiExpandChars(T1, T2).
cgiExpandChars([X | T1], [X | T2]) :- cgiExpandChars(T1, T2).

%
% cgiSend(List) or cgiSend(Term)
%
% Outputs the list of terms or the term (which may be a string, atom,
% number, etc. to the output file.
%
cgiSend([]) :-
   !,
   cgiWrite($\n$).
cgiSend(Stuff) :-
   list(Stuff), !,
   cgiSendList(Stuff).
cgiSend(Stuff) :-
   cgiWrite(Stuff),
   cgiWrite($\n$).

cgiSendList([]) :-
   cgiWrite($\n$).
cgiSendList([H | T]) :-
   cgiWrite(H),
   cgiSendList(T).

%
% cgiSendLocalHTMLFile(FileName)
%
% Outputs the contents of an existing HTML text file to the output file.
% You must output the header separately before calling this function.
% This allows you to intermix "canned" HTML with generated HTML
%
cgiSendLocalHTMLFile(FileName) :-
   fopen(Handle, FileName, r),
   repeat,
   read_string(Handle, Str),
   (Str == '!EOF' -> true ; cgiSend(Str)),
   Str == '!EOF',
   fclose(Handle).
cgiSendLocalHTMLFile(FileName) :-
   throw( [$Unable to open local file $, FileName, $\n$] ).

%
% cgiSendURL(URL)
%
% Tells the server to redirect to the specified URL
%
cgiSendURL(URL) :-
  cgiSend([$Location: $, URL]),
  cgiSend($$).

%
% cgiGetInputValues(Section, Functor, INIFile)
%
% Reads all the key value pairs in the specified section of the INIFile
% and asserts them in the form of fact(KeyAtom, ValueAtom) or for
% multivalued facts asserts fact(KeyAtom, ValueList).
%
cgiGetInputValues(Section, Functor, INIFile) :-
   getPrivateProfileSection(Section, KeyList, INIFile),
   cgiAssertFacts(Functor, KeyList).
cgiAssertFacts(_, []).
cgiAssertFacts(Functor, [Fact|Rest]) :-
   Fact =.. [_, Key, Value],
   cgiAddNewFact(Functor, Key, Value),
   cgiAssertFacts(Functor, Rest).

%
% cgiAddNewFact(Functor, Key Value)
%
% Asserts a new fact taking care to handle multivalued properly.
% Called by amzicgi.exe for standard CGI interface
%
cgiAddNewFact(fact, Key, Value) :-
   attr(Key, AttrList),
   member(multivalued, AttrList),
   cgiAddMulti(fact, Key, Value).
cgiAddNewFact(Functor, Key, Value) :-
   NewFact =.. [Functor, Key, Value],
   assertz(NewFact).

%
% Multivalued attributes have their values maintained in a list
% The first clause appends new values to the list.  The second
% clause creates the fact with a list containing the first value.
%
cgiAddMulti(Functor, Key, Value) :-
   CheckFact =.. [Functor, Key, CurVal],
   call(CheckFact), !,
   retract(CheckFact),
   NewFact =.. [Functor, Key, [Value | CurVal] ],
   assertz(NewFact).
cgiAddMulti(Functor, Key, Value) :-
   NewFact =.. [Functor, Key, [Value] ],
   assertz(NewFact).

%
% cgiAskField(Key, Prompt, Length)
%
% Outputs the Prompt into the current HTML output file with the
% input box of the specified Length.
%
cgiAskField(Key, Prompt, Length) :-
   cgiSend([Prompt, $<P>$]),
   (Length == none -> 
      cgiSend([$<INPUT NAME="$, Key, $"><P>$])
   ; 
      cgiSend([$<INPUT SIZE=$, Length, $ NAME="$, Key, $"><P>$]) 
   ).

%
% cgiAskYesNo(Key, Prompt)
%
% Outputs the Prompt and asks for a yes or no radio button to
% be checked.
%
cgiAskYesNo(Key, Prompt) :-
   cgiSend([Prompt, $<P>$]),
   cgiSend([$<P><INPUT NAME="$, Key, $" TYPE="RADIO" VALUE="Yes">Yes$]),
   cgiSend([$<P><INPUT NAME="$, Key, $" TYPE="RADIO" VALUE="No">No<P>$]),
   cgiSend([$<P><INPUT NAME="$, Key, $" TYPE="RADIO" VALUE="NA">Not applicable or Unknown<P>$]).

%
% cgiAskMenu(Key, Prompt, Choices)
%
% Outputs the Prompt and a drop-down menu containing the items in the
% Choices list.
%
cgiAskMenu(Key, Prompt, Choices, Type) :-
   cgiSend([Prompt, $<P>$]),
   (Type = multiple ->
      length(Choices, N),
      (N < 5 -> NN is N ; NN is N // 2),
      cgiSend([$<SELECT NAME="$, Key, $" $, $MULTIPLE SIZE=$, NN, $ >$])
   ;
      cgiSend([$<SELECT NAME="$, Key, $" >$])
   ),
   cgiAskMenuChoices(Choices),
   cgiSend($</SELECT><P>$).

cgiAskMenuChoices([]).
cgiAskMenuChoices([Choice | Rest]) :-
   cgiSend([$<OPTION>$, Choice]),
   cgiAskMenuChoices(Rest).

%
% cgiLog(Stuff)
%
% If logging is enabled (by the existence of a system fact named 'Log File')
% then write Stuff to it.  The first write causes the file to be opened,
% and all the system, cgi and facts to be written to it.  The file is
% closed automatically when the script finishes.
%
cgiLog(Stuff) :-
   var(Stuff), !.
cgiLog(tab(N)) :-
   system('Log File', LF), !,
   cgiOpenLog(LF, Handle),
   tab(Handle, N).
cgiLog(StuffList) :-
   list(StuffList),
   system('Log File', LF), !,
   cgiOpenLog(LF, Handle),
   cgiLogListWrite(Handle, StuffList).
cgiLog(Stuff) :-
   system('Log File', LF), !,
   cgiOpenLog(LF, Handle),
   write(Handle, Stuff).
cgiLog(Stuff) :- !.

cgiOpenLog(FileName, Handle) :-
   system('Log Handle', Handle), !.
cgiOpenLog(FileName, Handle) :-
   tell(FileName),
   write($<HTML><HEAD><TITLE>Logic-Base Debugging Trace</TITLE></HEAD>\n$),
   write($<BODY><BODY bgcolor=#FEFFDB text=#000000><H2>Logic-Base Debugging Trace</H2><PRE>\n$),
   cgiLogFactsWrite,
   nl,
   told,
   fopen(Handle, FileName, a),
   asserta(system('Log Handle', Handle)).

cgiLogFactsWrite :-
  cgiListing(system(_,_)),
  nl,
  cgiListing(cgi(_,_)),
  nl,
  cgiListing(fact(_,_)).
cgiLogFactsWrite.

cgiListing(Term) :-
  call(Term),
  string_term(Str, Term),
  string_list(Str, CharList),
  cgiExpandChars(CharList, CharListOut),
  string_list(StrOut, CharListOut),
  write(StrOut), nl,
  fail.
cgiListing(_).
  
cgiLogListWrite(_, []).
cgiLogListWrite(Handle, [H | T]) :-
   write(Handle, H),
   cgiLogListWrite(Handle, T).
   
cgiCloseLog :-
   system('Log Handle', Handle), !,
   write(Handle, $\n\n</PRE><P><HR>See the <A HREF="http://www.amzi.com">Amzi! Web Site</A> for more information and documentation</BODY></HTML>$),
   fclose(Handle).
cgiCloseLog.

%
% Utilities
%
append([], X, X).
append([A|X], Y, [A|Z]) :- append(X,Y,Z).

member(A, [A|_]).
member(A, [_|Z]) :- member(A, Z).

reverse(F, R) :-
	rev(F, []+R).

	rev([], R+R) :- !.
	rev([H|T], Ra+R) :-
		rev(T, [H|Ra]+R). 

length(L, N) :-length(L, 0, N).
  length([], N, N).
  length([_|Y], X, N) :-
    XX is X + 1,
    length(Y, XX, N).

%
% Extremely handy debugging tool.  Just put a ? in front of any Prolog predicate
% you want to trace (and enable logging).  See op definition at top of file.
%
bugwrite(X) :-
  cgiLog(X),
  cgiLog($\n$).

bugcallfail(X) :-
  bugwrite('CALL':X).
bugcallfail(X) :-
  bugwrite('FAIL':X),
  fail.

bugexitredo(X) :-
  bugwrite('EXIT':X).
bugexitredo(X) :-
  bugwrite('REDO':X),
  fail.

? X :- bug(X).

bug(X) :-
  bugcallfail(X),
  call(X),
  bugexitredo(X).

